"""
Definitions of metadata identifiers, types, and functions to operate on them.
"""
import re
from lrgasp import LrgaspException
from lrgasp.symEnum import SymEnum, auto

# fixed file names, without .gz
ENTRY_JSON = "entry.json"
EXPERIMENT_JSON = "experiment.json"
MODELS_GTF = "models.gtf"
READ_MODEL_MAP_TSV = "read_model_map.tsv"
EXPRESSION_TSV = "expression.tsv"

class Challenge(SymEnum):
    """Challenge identifiers"""
    iso_detect_ref = auto()
    iso_quant = auto()
    iso_detect_de_novo = auto()

class Sample(SymEnum):
    """LRGASP sample identifierd"""
    WTC11_Hs = auto()
    H1_DE_Hs = auto()
    ES_Mm = auto()
    Manatee = auto()

class LibraryPrep(SymEnum):
    """LRGASP library preperation methods"""
    Illumina_cDNA = auto()
    Pacbio_cDNA = auto()
    Pacbio_CapTrap = auto()
    ONT_cDNA = auto()
    ONT_CapTrap = auto()
    ONT_R2C2 = auto()
    ONT_dRNA = auto()
    Kitchen_Sink = auto()

class ExperimentType(SymEnum):
    "type of a experiment, assumed from Challenge type"
    model = auto()
    expression = auto()

class ExpressionUnits(SymEnum):
    "Units used in expression matrix"
    RPM = auto()
    RPKM = auto()
    FPKM = auto()
    TPM = auto()
    counts = auto()

def validate_symbolic_ident(ident):
    if not ident.isidentifier():
        raise LrgaspException(f"not a valid symbolic identifier: '{ident}'")

def validate_feature_ident(ident):
    if (not ident.isascii()) or (not ident.isprintable()) or re.search("\\s", ident) or (len(ident) == 0):
        raise LrgaspException(f"invalid feature identifier, must be composed of ASCII, printable, non-white-space characters: '{ident}'")

def validate_synapse_ident(ident):
    if not re.match("^syn[0-9]{4,30}$", ident):
        raise LrgaspException(f"not a valid Synapse identifier: '{ident}'")

def validate_entry_ident(entry_id):
    """check that an entry is prefix with one of the challenge ids, return the Challenge identifier that
    matches"""
    for ch in Challenge:
        if entry_id.startswith(str(ch) + '_'):
            return ch
    valid_pre = ", ".join([str(ch) + '_*' for ch in Challenge])
    raise LrgaspException("entry_id {} must be prefixed with a challenge id ({})".format(entry_id, valid_pre))

def challengeToExperimentType(challenge):
    assert isinstance(challenge, Challenge)
    if challenge is Challenge.iso_quant:
        return ExperimentType.expression
    else:
        return ExperimentType.model



"""
Experiment metadata parsing and validation.
"""
import os.path as osp
import json
from lrgasp import LrgaspException, gopen
from lrgasp.objDict import ObjDict
from lrgasp.defs import ExperimentType, Sample, ExpressionUnits, validate_symbolic_ident, EXPERIMENT_JSON
from lrgasp.metadata_validate import Field, check_from_defs, validate_url

fld_experiment_id = Field("experiment_id", validator=validate_symbolic_ident)
fld_experiment_type = Field("experiment_type", ExperimentType)
fld_description = Field("description")
fld_samples = Field("samples", list, element_dtype=Sample)
fld_data_files = Field("data_files", list, element_dtype=dict)
fld_units = Field("units", ExpressionUnits, optional=True)
fld_software = Field("software", list, element_dtype=dict)
fld_notes = Field("notes", allow_empty=True, optional=True)

experiment_fields = (
    fld_experiment_id,
    fld_experiment_type,
    fld_description,
    fld_samples,
    fld_data_files,
    fld_units,
    fld_software,
    fld_notes,
)

fld_acc = Field("acc", optional=True)
fld_url = Field("url", optional=True, validator=validate_url)

data_file_fields = (
    fld_acc,
    fld_url,
    fld_notes,
)

fld_name = Field("name")
fld_version = Field("version")
fld_url = Field("url", validator=validate_url)
fld_config = Field("url", allow_empty=True, optional=True)

experiment_software_fields = (
    fld_name,
    fld_version,
    fld_url,
    fld_config,
    fld_notes,
)

def data_file_validate(data_file):
    desc = "experiment.data_files"
    check_from_defs(desc, data_file_fields, data_file)
    if (fld_acc.name not in data_file) and (fld_url not in data_file):
        raise LrgaspException(f"{desc} must specify at least one of {fld_acc.name} or {fld_url.name}")

def experiment_validate(experiment):
    desc = "experiment"
    check_from_defs(desc, experiment_fields, experiment)

    for data_file in experiment.data_files:
        data_file_validate(data_file)

    if experiment.experiment_type is ExperimentType.model:
        if fld_units.name in experiment:
            raise LrgaspException(f"{desc} {experiment.experiment_type} must not specify {fld_units.name}")
    else:
        if fld_units.name not in experiment:
            raise LrgaspException(f"{desc} {experiment.experiment_type} must specify {fld_units.name}")

def load(experiment_json):
    """load and validate experiment metadata"""
    try:
        with gopen(experiment_json) as fh:
            experiment = json.load(fh, object_pairs_hook=ObjDict)
    except json.decoder.JSONDecodeError as ex:
        raise LrgaspException(f"parse of experiment metadata (JSON) failed: {experiment_json}") from ex
    try:
        experiment_validate(experiment)
    except LrgaspException as ex:
        raise LrgaspException(f"validation of experiment metadata failed: {experiment_json}") from ex
    return experiment

def load_from_entry(entry, experiment_id):
    """load and validation experiment metadata given a entry,
    add experiment_dir to metadata"""
    experiment_dir = osp.join(entry.entry_dir, experiment_id)
    experiment_json = osp.join(experiment_dir, EXPERIMENT_JSON)
    try:
        experiment = load(experiment_json)
    except (LrgaspException, FileNotFoundError, ValueError) as ex:
        raise LrgaspException(f"error parse metadata for entry {entry.entry_id}, experiment {experiment_id}: {experiment_json}") from ex
    experiment.experiment_dir = experiment_dir
    return experiment



"""Functions for validating metadata"""
import re
from collections import namedtuple
import validators.url
import validators.email
from lrgasp import LrgaspException

class Field(namedtuple("Field",
                       ("name", "dtype", "element_dtype", "allow_empty", "optional", "validator"))):
    """Specify information for basic validation of a field.
    - dtype is a Python type (str, SymEnum, int, etc).  If dtype is not str,
      int, or dict, it will be converted automatically.  If it is a
      list, each entry is processed using element_dtype if it is not None
    - validator is a callable to validate the type or each element if a list.
    """
    def __new__(cls, name, dtype=str, *, element_dtype=None, allow_empty=False, optional=False, validator=None):
        if (dtype is list) and (element_dtype is None):
            raise LrgaspException("list field {name} must specify element_dtype")
        return super().__new__(cls, name, dtype, element_dtype, allow_empty, optional, validator)

def _should_convert(dtype):
    return dtype not in (str, int, dict)

def _convert_type(desc, field, dtype, val):
    try:
        return dtype(val)
    except ValueError as ex:
        raise LrgaspException(f"{desc} field {field.name} is not a valid {dtype.__name__}: {val}") from ex

def _validate_value(desc, field, val):
    try:
        field.validator(val)
    except ValueError as ex:
        raise LrgaspException(f"{desc} field {field.name} is not valid: {val}") from ex

def _check_scalar(desc, field, val):
    # returns possibly converted value
    if _should_convert(field.dtype):
        val = _convert_type(desc, field, field.dtype, val)
    elif not isinstance(val, field.dtype):
        raise LrgaspException(f"{desc} field {field.name} must be a {field.dtype.__name__}")
    if field.validator is not None:
        _validate_value(desc, field, val)
    return val

def _check_list_element(desc, field, i, val):
    # returns possibly converted value
    if _should_convert(field.element_dtype):
        val = _convert_type(desc, field, field.element_dtype, val)
    elif not isinstance(val, field.element_dtype):
        raise LrgaspException(f"{desc} field {field.name}[{i}] must be a {field.element_dtype.__name__}")
    if field.validator is not None:
        _validate_value(desc, field, val)
    return val

def _check_list(desc, field, vals):
    new_vals = []
    if not isinstance(vals, list):
        raise LrgaspException(f"{desc} field {field.name} must be a list")
    for i in range(len(vals)):
        ival = _check_list_element(desc, field, i, vals[i])
        if ival in new_vals:
            print("@1", ival)
            print("@2", new_vals)
            raise LrgaspException(f"{desc} [{i}] field {field.name} duplicate value {ival}")
        new_vals.append(ival)
    return new_vals

def _check_present_field(desc, field, obj):
    val = getattr(obj, field.name)
    if (not field.allow_empty) and (len(val) == 0):
        raise LrgaspException(f"{desc} field {field.name} must be a non-empty {field.dtype.__name__}")
    if field.dtype is list:
        val = _check_list(desc, field, val)
    else:
        val = _check_scalar(desc, field, val)
    setattr(obj, field.name, val)

def _check_missing_field(desc, field, obj):
    if not field.optional:
        raise LrgaspException(f"{desc} field {field.name} is required")

def _check_field(desc, field, obj):
    if field.name in obj:
        _check_present_field(desc, field, obj)
    else:
        _check_missing_field(desc, field, obj)

def _check_for_unknown_fields(desc, fields, obj):
    fieldNames = frozenset([f.name for f in fields])
    bad = []
    for fld in obj.keys():
        if fld not in fieldNames:
            bad.append(fld)
    if len(bad) > 0:
        raise LrgaspException("{} unknown field name(s): {}".format(desc, ", ".join(bad)))

def validate_email(val):
    if validators.email(val) is not True:
        raise LrgaspException(f"invalid email address: {val}")

def validate_url(val):
    if validators.url(val, public=True) is not True:
        raise LrgaspException(f"invalid URL: {val}")

def validate_http_url(val):
    validate_url(val)
    if not (val.startswith("http:") or val.startswith("https:")):
        raise LrgaspException(f"must be an HTTP URL: {val}")

def validate_md5(val):
    "checks MD5 string"
    if not re.search("^[a-fA-F0-9]{32}$", val):
        raise LrgaspException(f"MD5 must be a 128-bit hexadecimal number: {val}")

def check_from_defs(desc, fields, obj):
    """basic check of fields given a definition table"""
    _check_for_unknown_fields(desc, fields, obj)
    for field in fields:
        _check_field(desc, field, obj)



"""
Load and validate model and expression entries
"""
import os.path as osp
from lrgasp import LrgaspException
from lrgasp.defs import ExperimentType, challengeToExperimentType, MODELS_GTF, READ_MODEL_MAP_TSV, EXPRESSION_TSV
from lrgasp import entry_metadata
from lrgasp import experiment_metadata
from lrgasp import model_data
from lrgasp import read_model_map_data
from lrgasp import expression_data

def _validate_trans_and_read_mapping(trans, read_model_map):
    if read_model_map.get_by_transcript_id(trans.transcript_id) is None:
        raise LrgaspException(f"transcript in models {trans.transcript_id} not in read-model_map")

def _validate_read_mapping_trans(transcript_id, models):
    if models.by_transcript_id.get(transcript_id) is None:
        raise LrgaspException(f"transcript in read_model_map {transcript_id} not in models")

def validate_model_and_read_mapping(models, read_model_map):
    # all model mapping must be in models
    for transcript_id in set([p.transcript_id for p in read_model_map]):
        # transcript might be None if specified as `*'
        if transcript_id is not None:
            _validate_read_mapping_trans(transcript_id, models)
    # all transcripts must be in model map
    for trans in models:
        _validate_trans_and_read_mapping(trans, read_model_map)

def _validate_model_experiment(entry, experiment):
    model_gtf = osp.join(experiment.experiment_dir, MODELS_GTF)
    map_file = osp.join(experiment.experiment_dir, READ_MODEL_MAP_TSV)
    try:
        models = model_data.load(model_gtf)
        read_model_map = read_model_map_data.load(map_file)
        validate_model_and_read_mapping(models, read_model_map)
    except Exception as ex:
        raise LrgaspException(f"entry {entry.entry_id} experiment {experiment.experiment_id} validation failed on {model_gtf} and {map_file}") from ex

def validate_expression_and_model(models, expression):
    # all expression matrix ids must be in models
    for row in expression.iterrows():
        if row[1].ID not in models.by_transcript_id:
            raise LrgaspException(f"expression matrix ID {row[1].ID} not found in models")

def _validate_expression_experiment(entry, experiment):
    model_gtf = osp.join(experiment.experiment_dir, MODELS_GTF)
    expression_tsv = osp.join(experiment.experiment_dir, EXPRESSION_TSV)
    try:
        models = model_data.load(model_gtf)
        expression = expression_data.load(expression_tsv)
        validate_expression_and_model(models, expression)
    except Exception as ex:
        raise LrgaspException(f"entry {entry.entry_id} experiment {experiment.experiment_id} validation failed on {model_gtf} and {expression_tsv}") from ex

def _validate_experiment(entry, experiment_id):
    experiment = experiment_metadata.load_from_entry(entry, experiment_id)
    experiment_type = challengeToExperimentType(entry.challenge_id)
    if experiment.experiment_type is not experiment_type:
        raise LrgaspException(f"entry {entry.entry_id} challenge {entry.challenge_id} does not consistent with experiment {experiment_id} type {experiment_type}")
    if experiment_type == ExperimentType.model:
        _validate_model_experiment(entry, experiment)
    else:
        _validate_expression_experiment(entry, experiment)

def entry_data_validate(entry_dir, restrict_experiment_id=None):
    """load and validate all metadata and data files for an entry, ensuring
    consistency.  Optionally restricted to one experiment for speed"""
    entry = entry_metadata.load_dir(entry_dir)

    if restrict_experiment_id is not None:
        if restrict_experiment_id not in entry.experiment_ids:
            raise LrgaspException(f"entry {entry.entry_id} does not contain experiment {restrict_experiment_id}")
        experiment_ids = [restrict_experiment_id]
    else:
        experiment_ids = entry.experiment_ids

    for experiment_id in experiment_ids:
        _validate_experiment(entry, experiment_id)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import setuptools

requirements = [
    "gtfparse>=1.2.1",
    "validators>=0.18.2"
]

setuptools.setup(
    name = 'lrgasp-tools',
    version = '0.7.0',
    description = "LRGASP tools",
    long_description = "LRGASP tools for submission",
    author = "Mark Diekhans",
    author_email = 'markd@ucsc.edu',
    url = 'https://github.com/LRGASP/lrgasp-submissions',
    scripts=[
        'bin/lrgasp-validate-entry',
        'bin/lrgasp-validate-entry-metadata',
        'bin/lrgasp-validate-experiment-metadata',
        'bin/lrgasp-validate-expression-matrix',
        'bin/lrgasp-validate-models',
        'bin/lrgasp-validate-read-model-map',
    ],
    packages = [
        'lrgasp',
    ],
    package_dir = {'': 'lib'},
    include_package_data = True,
    install_requires = requirements,
    license = "MIT",
    zip_safe = True,
    keywords = ['Bioinformatics', 'genomics', 'transcriptomics'],
    classifiers = [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    python_requires='>=3.7',
)



# Copyright 2006-2014 Mark Diekhans

from enum import Enum, EnumMeta
from functools import total_ordering

# FIXME: should really be built like other IntEnum, etc

from enum import auto


SymEnum = None  # class defined after use


class SymEnumValue(object):
    "Class used to define SymEnum member that have additional attributes."
    __slots__ = ("value", "externalName")

    def __init__(self, value, externalName=None):
        self.value = value
        self.externalName = externalName


class _SysEnumExternalNameMap(object):
    "Mapping between internal and external member names"
    __slots__ = ("internalToExternal", "externalToInternal")

    def __init__(self):
        self.internalToExternal = {}
        self.externalToInternal = {}

    def add(self, internalName, externalName):
        self.internalToExternal[internalName] = externalName
        self.externalToInternal[externalName] = internalName

    def toExternalName(self, internalName):
        "return name unchanged if no mapping"
        return self.internalToExternal.get(internalName, internalName)

    def toInternalName(self, externalName):
        "return name unchanged if no mapping"
        return self.externalToInternal.get(externalName, externalName)

    def __str__(self):
        return "intToExt={} extToInt={}".format(self.internalToExternal, self.externalToInternal)


class SymEnumMeta(EnumMeta):
    """metaclass for SysEnumMeta that implements looking up of singleton members
    by string name."""

    @classmethod
    def _symEnumValueUpdate(cls, classdict, name, externalNameMap):
        """record info about a member specified with SymEnum and update value in classdict
        to be actual value rather than SymEnumValue"""
        symValue = classdict[name]
        # under python3, enum does some sanity check, must remove from
        # _EnumDict before reinserting
        del classdict._member_names[classdict._member_names.index(name)]
        del classdict[name]
        classdict[name] = symValue.value
        externalNameMap.add(name, symValue.externalName)

    @classmethod
    def _buildExternalNameMap(cls, classdict):
        "fill in external name map"
        externalNameMap = classdict["__externalNameMap__"] = _SysEnumExternalNameMap()
        for name in list(classdict.keys()):  # MUST COPY, as modifying classdict
            if isinstance(classdict[name], SymEnumValue):
                cls._symEnumValueUpdate(classdict, name, externalNameMap)

    def __new__(metacls, clsname, bases, classdict):
        "updates class fields defined as SymEnumValue to register external names"
        if SymEnum in bases:
            metacls._buildExternalNameMap(classdict)
        return super(SymEnumMeta, metacls).__new__(metacls, clsname, bases, classdict)

    @staticmethod
    def _lookUpByStr(cls, value):
        # map string name to instance, check for external name
        member = cls._member_map_.get(cls.__externalNameMap__.toInternalName(value))
        if member is None:
            member = cls._member_map_.get(cls.__externalNameMap__.toExternalName(value))
        if member is None:
            raise ValueError("'{}' is not a member, external name, or alias of {}".format(value, cls.__name__))
        else:
            return member

    def __call__(cls, value, names=None, module=None, typ=None):
        "look up a value object, either by name or value"
        if (names is None) and isinstance(value, str):
            return SymEnumMeta._lookUpByStr(cls, value)
        else:
            return EnumMeta.__call__(cls, value, names, module=module, type=typ)


@total_ordering
class SymEnumMixin(object):
    """Mixin that adds comparisons and other functions for SymEnum"""

    def __hash__(self):
        return hash(self.value)

    def __eq__(self, other):
        if isinstance(other, SymEnum):
            return self.value == other.value
        else:
            return self.value == other

    def __lt__(self, other):
        if isinstance(other, SymEnum):
            return self.value < other.value
        else:
            return self.value < other

    def __reduce_ex__(self, proto):
        return self.__class__, (self.value, )


class SymEnum(SymEnumMixin, Enum, metaclass=(SymEnumMeta)):
    """
    Metaclass for symbolic enumerations.  These are easily converted between
    string values and Enum objects.  This support construction from string
    values and str() returns value without class name.  Aliases can be
    added using the Enum approach of:
        name = 1
        namealias = 1

    The functional API works as with Enum. The auto() method of
    field initialization works with limited functionality on Python2.

    To handle string values that are not valid Python member names, an external
    name maybe associated with a field using a SymEnumValue object
        utr5 = SymEnumValue(1, "5'UTR")

    Either field name or external name maybe used to obtain a value.  The external
    name is returned with str().

    Instances of the enumerations are obtained by either string name or int
    value:
       SymEnum(strName)
       SymEnum(intVal)

    To use as an argument type in argparser
       type=Color, choices=Color
    """

    def __str__(self):
        return self.__externalNameMap__.toExternalName(self.name)

    def __format__(self, fmtspec):
        return format(str(self), fmtspec)


__all__ = (SymEnumValue.__name__, SymEnum.__name__, auto.__name__)
